/**
 ** A Virtual Host Monitor
 **
 ** Copyright 2009 Gary Wuertz gary@issiweb.com
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <iostream>

#include <errno.h>
#include <getopt.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <ncurses.h>

#include "monitor.h"

using namespace std;
/**
 * Diagnostic to display field associations
 */
void topvhost::diagMap(unsigned nidx)
{
   map<int,int>::iterator idb;
   string sep="";
   unsigned j;

   cout << "Output (Needs): '" << this->rfmt << "'" << endl;
   for(j=0;j<this->needs.size();j++) {
      cout << sep << j << " = " << this->needs[j];
      sep = ", ";
      }
   cout << endl << "Input (Gets): '" << this->logfmt << "'" << endl;
   sep = "";
   for(j=0;j<this->gets.size();j++) {
      cout << sep << j << " = " << this->gets[j];
      sep = ", ";
      }
   cout << endl << "Transfer Map (Gets => Needs)" << endl;
   sep = "";
   for(idb = this->xfer.begin();idb != this->xfer.end();++idb) {
      cout << sep << idb->first << " => " << idb->second;
      sep = ", ";
      }
   cout << endl << "Domain name @" << nidx << endl;
}
/**
 * Format page header using a sprintf like specification using the place holders:
 *
 *  %H    = hour
 *  %I    = bytes read
 *  %M    = minute
 *  %S    = second
 *  %Y    = year
 *  %d    = day
 *  %m    = month
 *  %o    = total hits
 *  %n    = total vhosts
 *  %t,%T = elapsed time (hh:mm:ss or n)
 */
void topvhost::getHeader()
{
   char    buf[128];
   time_t  seconds = time(NULL);
   struct  tm *ts  = localtime(&seconds);
   string  token;
   int     ch,val=0;long lval=0;
   string  stype,sval;
   string::iterator fmt = this->hfmt.begin();

   if (this->passes++==0)
      this->start = seconds;
   this->tline = "";
   while( fmt != this->hfmt.end() )
      if ((ch = *fmt++)=='%') {
         stype = "d";
         token = ch;
         if (strchr("-+#0 ",*fmt)!=NULL)           // flags
            token += *fmt++;
         while(strchr("0123456789.",*fmt)!=NULL)   // width - precision
            token += *fmt++;
         switch(*fmt++) {
            case 'H':   val = ts->tm_hour;    break;
            case 'I':
               if (token=="%#") {
                  int factors[] = {1, ' ', 1024, 'K', 1024*1024, 'M', 1024*1024*1024, 'G'};
                  int i;
                  for (i=0;i<=2 && this->meter>=static_cast<long>(factors[i+2]);i+=2) ;
                  snprintf(buf,sizeof(buf),"%6.3g%c",
                     static_cast<float>(this->meter)/factors[i], factors[i+1]);
                  stype="s";
                  break;
                  }
               lval = this->meter;
               stype = "lu";
               break;
            case 'M':   val = ts->tm_min;     break;
            case 'S':   val = ts->tm_sec;     break;
            case 'T':
               val = static_cast<int>(seconds - this->start);
               break;
            case 'Y':   val = ts->tm_year;    break;
            case 'd':   val = ts->tm_mday;    break;
            case 'm':   val = ts->tm_mon+1;   break;
            case 'n':   val = this->total;    break;
            case 'o':   val = this->hits;     break;
            case 't':
               val = static_cast<int>(seconds - this->start);
               snprintf(buf,sizeof(buf),"%02d:%02d:%02d",(val/3600)%60,(val/60)%60,val%60);
               stype = "s";
            }
         token += stype;
         if (stype=="d")
            snprintf(buf,sizeof(buf),token.c_str(), val);
         else if (stype=="lu")
            snprintf(buf,sizeof(buf),token.c_str(), lval);
         this->tline += buf;
         }
      else this->tline += ch;
}

/**
 * Local control structure
 */
static topvhost info;
/**
 * Handle KILL and INT signals
 */
static void catchsig(int s)
{
   info.exit = s;
}
/**
 * Handle window resize signal
 */
static void catchwinch(int s)
{
   info.resize = true;
}
/**
 * Main loop
 */
int main(int argc, char **argv)
{
   static char * cmds[] = {
      "d", "init",   "0", "Enable diagnostics",
      "f", "file",   "1", "Configuration file",
      "i", "init",   "0", "Parse logs at startup",
      "s", "select", "1", "Select Configuration",
      "v", "verify", "0", "Verify Configuration",
      "h", "help",   "0", "This help"
      };
   static struct monitorOptions options = {"", "", false, false };
   int nopts = sizeof(cmds)/(4*sizeof(char *));
   struct option* long_options = new option[nopts];
   string short_options;
   int c,i,j;
   Monitor mon;
   vhost *inst = NULL;
   WINDOW *win;

   for(i=j=0;i<nopts;i++,j+=4) {
      long_options[i].name      = cmds[j+1];
      long_options[i].has_arg   = atoi(cmds[j+2]);
      long_options[i].flag      = NULL;
      long_options[i].val       = cmds[j][0];
      short_options += cmds[j];
      if (long_options[i].has_arg!=0) short_options += ":";
      }
   do {
      c = getopt_long (argc, argv, short_options.c_str(), long_options, NULL);
      switch(c) {
         case 'd':
            info.debug = 1;
            break;
         case 'f':
            options.configFile = optarg;
            break;
         case 'i':
            options.init   = true;
            break;
         case 's':
            options.configSection = optarg;
            break;
         case 'h':
            fprintf(stderr, "\nUsage: %s [options]\n\n", argv[0]);
            fprintf(stderr, "Monitor Virtual Hosts.\n");
            fprintf(stderr, "\n" "Options:\n");
            for(i=0;long_options[i].name!=NULL;i++)
               fprintf(stderr,"--%-10s, -%c %s %s\n",
                  long_options[i].name, long_options[i].val,
                  long_options[i].has_arg? "[]":"  ",cmds[i*4+3]);
            fprintf(stderr, "\n");
            exit(1);
         case 'v':
            options.verify = true;
         }
      } while (c!=-1);
   if (!mon.init(&info, options)) {
      cout << mon.problem << endl;
      return -1;
      }
   if (options.verify) {
      inst = mon.getHosts(&info);
      cout << "Logs Monitored" << endl;
      for(i=0;i<info.total;i++) {
         cout << inst[i].filepath() << ":" << endl;
         inst[i].setString(&info);
         cout << info.tline << endl;
         }
      mon.quit();
      exit(1);
      }
   signal(SIGINT, &catchsig);
   signal(SIGKILL, &catchsig);
   signal(SIGWINCH, &catchwinch);
   info.paint = true;
   win  = initscr();
   getmaxyx(win,info.maxy,info.maxx);
   cbreak();
   nodelay(win,true);
   while(info.exit==0) {              // exit on signal or error
      switch(getch()) {
         case 'd':
            info.debug = !info.debug;
            break;
         case 'h':   case 'H':
            info.order = topvhost::BY_HITS;
            info.paint = true;
            break;
         case 'n':   case 'N':
            info.order = topvhost::BY_NAME;
            info.paint = true;
            break;
         case 'q':   case 'Q':
            info.exit = -2;
            break;
         case 'r':   case 'R':
            mon.refresh();
            info.paint = true;
            break;
         case 't':   case 'T':
            info.order = topvhost::BY_ACCESS;
            info.paint = true;
            break;
         }
      if (info.resize) {
         info.resize = 0;
         endwin();
         refresh();
         getmaxyx(win,info.maxy,info.maxx);
         }
      else if (info.paint)
         info.paint = false;
      else if (!mon.update(&info)) {
         info.exit = -1;
         break;
         }
      inst = mon.getHosts(&info);
      info.getHeader();
      mvprintw(0,0,info.tline.c_str());
      for(i=0;i<info.visible;i++) {
         inst[i].setString(&info);
         mvprintw(i+info.hdrsz,0,info.tline.c_str());
         clrtoeol();
         }
      mvprintw(1,0,"|");
      clrtoeol();
      refresh();
      sleep(1);
      }
   endwin();
   mon.quit();
   switch(info.exit) {
      case -2:
         cout << "Quit" << endl;
         break;
      case -1:
         cout << "Update " << mon.problem << endl;
         break;
      default:
         cout << "Caught signal " << info.exit << endl;
      }
   return 0;
}
